"""
Jarvis WRP Runtime Repository

Repository containing Chimaera runtime deployment packages for Jarvis-CD.
This repository provides Service packages for distributed deployment and 
management of the Chimaera runtime system.
"""


##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Defines wx.Frame subclasses which are the foundation of the various windows
displayed by HDFCompass.

Much of the common functionality (e.g. "Open File..." menu item) is implemented
here.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
from datetime import date

import wx
import wx.richtext as rtc
from wx.lib.pubsub import pub
from wx.adv import AboutDialogInfo, AboutBox

log = logging.getLogger(__name__)

from .info import InfoPanel

ID_OPEN_RESOURCE = wx.ID_ANY
ID_CLOSE_FILE = wx.ID_ANY
ID_PLUGIN_INFO = wx.ID_ANY

MAX_RECENT_FILES = 8

from hdf_compass import compass_model
from hdf_compass.utils import __version__, is_darwin, path2url
from .events import CompassOpenEvent

open_frames = 0  # count the open frames


class BaseFrame(wx.Frame):
    """
    Base class for all frames used in HDF Compass.

    Implements common menus including File and Help, and handles their
    events.

    When implementing a new viewer window, you should instead inherit from
    BaseFrame (below), which adds a left-hand side information panel, and
    participates in the reference counting that automatically shows the
    initial window when all other frames are closed.
    """

    icon_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), 'icons'))
    open_frames = 0  # count the number of frames

    last_open_path = os.getcwd()

    def __init__(self, **kwds):
        """ Create a new base frame """
        log.debug(self.__class__.__name__)
        # Handle pos=None case for wxPython compatibility
        if 'pos' in kwds and kwds['pos'] is None:
            kwds['pos'] = wx.DefaultPosition
        super(BaseFrame, self).__init__(None, **kwds)

        # Set icon
        icon_32 = wx.Icon()
        icon_32.CopyFromBitmap(wx.Bitmap(os.path.join(self.icon_folder, "favicon_32.png"), wx.BITMAP_TYPE_ANY))
        icon_48 = wx.Icon()
        icon_48.CopyFromBitmap(wx.Bitmap(os.path.join(self.icon_folder, "favicon_48.png"), wx.BITMAP_TYPE_ANY))
        
        # Create icon bundle and add icons
        icon_bundle = wx.IconBundle()
        icon_bundle.AddIcon(icon_32)
        icon_bundle.AddIcon(icon_48)
        self.SetIcons(icon_bundle)

        # Create menu bar
        self.menu_bar = wx.MenuBar()
        self.SetMenuBar(self.menu_bar)

        # Create file menu
        file_menu = wx.Menu()
        file_menu.Append(wx.ID_OPEN, "&Open...\tCtrl+O")
        file_menu.Append(wx.ID_SAVE, "&Save\tCtrl+S")
        file_menu.Append(wx.ID_SAVEAS, "Save &As...\tCtrl+Shift+S")
        file_menu.AppendSeparator()
        file_menu.Append(wx.ID_CLOSE, "&Close\tCtrl+W")
        file_menu.Append(wx.ID_EXIT, "E&xit\tAlt+F4")
        self.menu_bar.Append(file_menu, "&File")

        # Create edit menu
        edit_menu = wx.Menu()
        edit_menu.Append(wx.ID_COPY, "&Copy\tCtrl+C")
        edit_menu.Append(wx.ID_PASTE, "&Paste\tCtrl+V")
        self.menu_bar.Append(edit_menu, "&Edit")

        # Create view menu
        view_menu = wx.Menu()
        view_menu.Append(wx.ID_ZOOM_IN, "Zoom &In\tCtrl++")
        view_menu.Append(wx.ID_ZOOM_OUT, "Zoom &Out\tCtrl+-")
        view_menu.Append(wx.ID_ZOOM_FIT, "Zoom &Fit\tCtrl+0")
        self.menu_bar.Append(view_menu, "&View")

        # Create help menu
        help_menu = wx.Menu()
        help_menu.Append(wx.ID_ABOUT, "&About")
        self.menu_bar.Append(help_menu, "&Help")

        # Create toolbar
        self.toolbar = self.CreateToolBar()
        self.toolbar.AddTool(wx.ID_OPEN, "Open", wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN))
        self.toolbar.AddTool(wx.ID_SAVE, "Save", wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE))
        self.toolbar.AddTool(wx.ID_COPY, "Copy", wx.ArtProvider.GetBitmap(wx.ART_COPY))
        self.toolbar.AddTool(wx.ID_PASTE, "Paste", wx.ArtProvider.GetBitmap(wx.ART_PASTE))
        self.toolbar.AddTool(wx.ID_ZOOM_IN, "Zoom In", wx.ArtProvider.GetBitmap(wx.ART_PLUS))
        self.toolbar.AddTool(wx.ID_ZOOM_OUT, "Zoom Out", wx.ArtProvider.GetBitmap(wx.ART_MINUS))
        self.toolbar.AddTool(wx.ID_ZOOM_FIT, "Zoom Fit", wx.ArtProvider.GetBitmap(wx.ART_FIND))
        self.toolbar.Realize()

        # Bind events
        self.Bind(wx.EVT_MENU, self.on_open, id=wx.ID_OPEN)
        self.Bind(wx.EVT_MENU, self.on_save, id=wx.ID_SAVE)
        self.Bind(wx.EVT_MENU, self.on_save_as, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_MENU, self.on_close, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_MENU, self.on_exit, id=wx.ID_EXIT)
        self.Bind(wx.EVT_MENU, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_MENU, self.on_paste, id=wx.ID_PASTE)
        self.Bind(wx.EVT_MENU, self.on_zoom_in, id=wx.ID_ZOOM_IN)
        self.Bind(wx.EVT_MENU, self.on_zoom_out, id=wx.ID_ZOOM_OUT)
        self.Bind(wx.EVT_MENU, self.on_zoom_fit, id=wx.ID_ZOOM_FIT)
        self.Bind(wx.EVT_MENU, self.on_about, id=wx.ID_ABOUT)

    def on_open(self, evt):
        """ Handle open menu item """
        with wx.FileDialog(self, "Open file", wildcard="All files (*.*)|*.*",
                          style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                path = dlg.GetPath()
                # Convert file path to URL format and open directly
                url = 'file://' + os.path.abspath(path)
                from .viewer import open_store
                open_store(url)

    def on_save(self, evt):
        """ Handle save menu item """
        pass

    def on_save_as(self, evt):
        """ Handle save as menu item """
        pass

    def on_close(self, evt):
        """ Handle close menu item """
        self.Close()

    def on_exit(self, evt):
        """ Handle exit menu item """
        wx.Exit()

    def on_copy(self, evt):
        """ Handle copy menu item """
        pass

    def on_paste(self, evt):
        """ Handle paste menu item """
        pass

    def on_zoom_in(self, evt):
        """ Handle zoom in menu item """
        pass

    def on_zoom_out(self, evt):
        """ Handle zoom out menu item """
        pass

    def on_zoom_fit(self, evt):
        """ Handle zoom fit menu item """
        pass

    def on_about(self, evt):
        """ Handle about menu item """
        wx.MessageBox("HDF Compass Viewer\n\nA viewer for HDF files",
                     "About HDF Compass", wx.OK | wx.ICON_INFORMATION)

    def add_menu(self, menu, title):
        """ Add a menu to the menu bar """
        menubar = self.GetMenuBar()
        if not menubar:
            menubar = wx.MenuBar()
            self.SetMenuBar(menubar)
        menubar.Append(menu, title)


class InitFrame(BaseFrame):
    """ Frame displayed when the application starts up.

    This includes the menu bar provided by TopFrame.  On the Mac, although it
    still exists (to prevent the application from exiting), the frame
    is typically not shown.
    """

    def __init__(self):
        style = wx.DEFAULT_FRAME_STYLE & (~wx.RESIZE_BORDER) & (~wx.MAXIMIZE_BOX)
        title = "HDF Compass"
        super(InitFrame, self).__init__(size=(552, 247), title=title, style=style)

        data = wx.Bitmap(os.path.join(self.icon_folder, "logo.png"), wx.BITMAP_TYPE_ANY)
        bmp = wx.StaticBitmap(self, wx.ID_ANY, data)

        # The init frame isn't visible on Mac, so there shouldn't be an
        # option to close it.  "Quit" does the same thing.
        if is_darwin:
            mb = self.GetMenuBar()
            mu = mb.GetMenu(0)
            mu.Enable(wx.ID_CLOSE, False)
        self.Center()


class NodeFrame(BaseFrame):
    """ Base class for any frame which displays a Node instance.

    Provides a "Close file" menu item and manages open data stores.

    Has three attributes of note:

    .node:  Settable Node instance to display
    .info:  Read-only InfoPanel instance (left-hand sidebar)
    .view:  Settable wx.Panel instance for the right-hand view.

    In order to coordinate file-close events across multiple frames,
    a reference-counting system is used.  When a new frame that uses a store
    is created, that store's reference count (in cls._stores) is incremented.
    When the frame is closed, the store's count is decremented.  

    When the reference count reaches 0 or the "Close File" is selected from the
    menu, the store is closed and a pubsub notification is sent out to all
    other frames.  They check to see if their .node.store's are valid, and
    if not, close themselves.
    """

    # --- Store reference-counting methods ------------------------------------

    _stores = {}

    @classmethod
    def _incref(cls, store):
        """ Record that a client is using the specified store. """
        try:
            cls._stores[store] += 1
        except KeyError:
            cls._stores[store] = 1

    @classmethod
    def _decref(cls, store):
        """ Record that a client is finished using the specified store. """
        try:
            val = cls._stores[store]
            if val == 1:
                cls._close(store)
                del cls._stores[store]
            else:
                cls._stores[store] = val - 1
        except KeyError:
            pass

    @classmethod
    def _close(cls, store):
        """ Manually close the store, and broadcast a pubsub notification. """
        cls._stores.pop(store, None)
        store.close()
        pub.sendMessage('store.close')

    # --- End store reference-counting ----------------------------------------

    @property
    def info(self):
        """ The InfoPanel object used for the left-hand sidebar. """
        return self.__info

    @property
    def node(self):
        """ Node instance displayed by the frame. """
        return self.__node

    @node.setter
    def node(self, newnode):
        self.__node = newnode

    @property
    def view(self):
        """ Right-hand view """
        return self.__view

    @view.setter
    def view(self, window):
        if self.__view is None:
            self.__sizer.Add(window, 1, wx.EXPAND)
        else:
            self.__sizer.Detach(self.__view)
            self.__view.Destroy()
            self.__sizer.Add(window, 1, wx.EXPAND)
        self.__view = window
        self.Layout()

    def __init__(self, node, **kwds):
        """ Constructor.  Keywords are passed on to wx.Frame.

        node:   The compass_model.Node instance to display.
        """

        super(NodeFrame, self).__init__(**kwds)

        # Enable the "Close File" menu entry if it exists
        try:
            menubar = self.GetMenuBar()
            if menubar and menubar.GetMenuCount() > 0:
                fm = menubar.GetMenu(0)
                if fm and fm.FindItem(ID_CLOSE_FILE) != wx.NOT_FOUND:
                    fm.Enable(ID_CLOSE_FILE, True)
        except wx.wxAssertionError:
            # Menu item doesn't exist, ignore
            pass

        # Create the "window" menu to hold "Reopen As" items.
        wm = wx.Menu()

        # Determine a list of handlers which can understand this object.
        # We exclude the default handler, "Unknown", as it can't do anything.
        # See also container/list.py.
        handlers = [x for x in node.store.gethandlers(node.key) if x != compass_model.Unknown]

        # This will map menu IDs -> Node subclass handlers
        self._menu_handlers = {}

        # Note there's guaranteed to be at least one entry: the class
        # being used for the current frame!
        for h in handlers:
            id_ = wx.ID_ANY
            self._menu_handlers[id_] = h
            wm.Append(id_, "Reopen as " + h.class_kind)
            self.Bind(wx.EVT_MENU, self.on_menu_reopen, id=id_)

        self.GetMenuBar().Insert(1, wm, "&Window")

        self.__node = node
        self.__view = None
        self.__info = InfoPanel(self)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__info, 0, wx.EXPAND)
        self.SetSizer(self.__sizer)

        self.info.display(node)

        # Modify existing File menu to add node-specific items
        menubar = self.GetMenuBar()
        if menubar and menubar.GetMenuCount() > 0:
            file_menu = menubar.GetMenu(0)  # File menu is typically the first menu
            if file_menu:
                # Add separator and node-specific menu items
                file_menu.AppendSeparator()
                file_menu.Append(ID_CLOSE_FILE, "Close &File\tCtrl+Shift+W")
        else:
            # Fallback: create File menu if none exists
            file_menu = wx.Menu()
            file_menu.Append(wx.ID_CLOSE, "&Close Window\tCtrl+W")
            file_menu.Append(ID_CLOSE_FILE, "Close &File\tCtrl+Shift+W")
            file_menu.AppendSeparator()
            file_menu.Append(wx.ID_EXIT, "E&xit\tAlt+F4")
            self.add_menu(file_menu, "&File")

        self.Bind(wx.EVT_CLOSE, self.on_close_evt)
        self.Bind(wx.EVT_MENU, self.on_close_window, id=wx.ID_CLOSE)
        self.Bind(wx.EVT_MENU, self.on_menu_closefile, id=ID_CLOSE_FILE)
        self.Bind(wx.EVT_MENU, self.on_exit, id=wx.ID_EXIT)

        self._incref(node.store)
        pub.subscribe(self.on_notification_closefile, 'store.close')

    def on_notification_closefile(self):
        """ Pubsub notification that a file (any file) has been closed """
        try:
            # Check if frame still exists and store is invalid
            if hasattr(self, 'node') and hasattr(self.node, 'store') and not self.node.store.valid:
                # Check if the frame hasn't been destroyed already
                if not self.IsBeingDeleted():
                    self.Destroy()
        except RuntimeError as e:
            # Frame was already destroyed
            log.debug(f"Frame already destroyed: {e}")
        except Exception as e:
            log.exception(f"Error in on_notification_closefile: {e}")

    def on_close_evt(self, evt):
        """ Window is about to be closed """
        self._decref(self.node.store)
        evt.Skip()

    def on_close_window(self, evt):
        """ Close Window menu item activated - close just this window """
        self.Close()

    def on_exit(self, evt):
        """ Exit menu item activated - exit application """
        wx.Exit()

    def on_menu_closefile(self, evt):
        """ "Close File" menu item activated.

        Note we rely on the pubsub message (above) to actually close the frame.
        """
        self._close(self.node.store)

    def on_menu_reopen(self, evt):
        """ Called when one of the "Reopen As" menu items is clicked """

        # The "Reopen As" submenu ID
        id_ = evt.GetId()

        # Present node
        node_being_opened = self.node

        # The requested Node subclass to instantiate.
        h = self._menu_handlers[id_]

        log.debug('opening: %s %s' % (node_being_opened.store, node_being_opened.key))
        # Brand new Node instance of the requested type
        node_new = h(node_being_opened.store, node_being_opened.key)

        # Send off a request for it to be opened in the appropriate viewer
        # Post it directly to the App, or Container will intercept it!
        pos = wx.GetTopLevelParent(self).GetPosition()
        wx.PostEvent(wx.GetApp(), CompassOpenEvent(node_new, pos=pos))


class PluginInfoFrame(wx.Frame):
    icon_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), 'icons'))

    def __init__(self, parent):
        # make that the plugin info is displayed in the middle of the screen
        frame_w = 320
        frame_h = 250
        x = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_X) // 2 - frame_w // 2
        y = wx.SystemSettings.GetMetric(wx.SYS_SCREEN_Y) // 2 - frame_h // 2
        wx.Frame.__init__(self, parent, title="Plugin Info", pos=(x, y), size=(frame_w, frame_h))

        # Frame icon
        ib = wx.IconBundle()
        icon_32 = wx.Icon()
        icon_32.CopyFromBitmap(wx.Bitmap(os.path.join(self.icon_folder, "favicon_32.png"), wx.BITMAP_TYPE_ANY))
        ib.AddIcon(icon_32)
        icon_48 = wx.Icon()
        icon_48.CopyFromBitmap(wx.Bitmap(os.path.join(self.icon_folder, "favicon_48.png"), wx.BITMAP_TYPE_ANY))
        ib.AddIcon(icon_48)
        self.SetIcons(ib)

        p = wx.Panel(self)
        nb = wx.Notebook(p)

        for store in compass_model.get_stores():
            try:
                # log.debug(store.plugin_name())
                # log.debug(store.plugin_description())

                pnl = wx.Panel(nb)
                t = rtc.RichTextCtrl(pnl, -1, style=wx.TE_READONLY)
                t.BeginFontSize(9)
                t.BeginAlignment(wx.TEXT_ALIGNMENT_CENTRE)
                t.BeginBold()
                t.WriteText("Name: ")
                t.EndBold()
                t.BeginItalic()
                t.WriteText(store.plugin_name())
                t.EndItalic()
                t.Newline()
                t.Newline()
                t.BeginBold()
                t.WriteText("Description")
                t.EndBold()
                t.Newline()
                t.BeginItalic()
                t.WriteText(store.plugin_description())
                t.EndItalic()
                t.Newline()

                # store.plugin_description(), style=wx.TE_MULTILINE|wx.TE_READONLY|wx.TE_CENTER)
                szr = wx.BoxSizer()
                szr.Add(t, 1, wx.ALL | wx.EXPAND, 5)
                pnl.SetSizer(szr)
                nb.AddPage(pnl, store.plugin_name())

            except NotImplementedError:
                # skip not implemented plugin name/description
                log.debug("Not implemented name/description for %s" % store)

        sizer = wx.BoxSizer()
        sizer.Add(nb, 1, wx.ALL | wx.EXPAND, 3)
        p.SetSizer(sizer)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
HDF Compass plugin for viewing CSV files.

Provides support for reading and viewing CSV files as array data.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path as op
import pandas as pd
import numpy as np
import logging

log = logging.getLogger(__name__)

from hdf_compass import compass_model
from hdf_compass.utils import url2path


class StructuredSlice:
    """
    A wrapper for numpy array slices that supports both integer indexing and structured field access.
    This is needed for LRUTileCache compatibility with structured arrays.
    """
    def __init__(self, data, columns):
        self.data = data
        self.columns = columns
        self.shape = data.shape

    def __getitem__(self, index):
        # Handle tuple indexing (e.g., (0,) from LRUTileCache)
        if isinstance(index, tuple) and len(index) == 1:
            index = index[0]

        # Handle string field name access (for plotting)
        if isinstance(index, str):
            field_name = index
            for i, col in enumerate(self.columns):
                clean_name = str(col).replace("'", "").replace("[", "_").replace("]", "_").replace(" ", "_")
                if clean_name == field_name:
                    return self.data[:, i]
            raise KeyError(f"Field {field_name} not found")

        if isinstance(index, int):
            # Return a structured record-like object for single row access
            row_data = self.data[index, :]
            record = {}
            for i, col in enumerate(self.columns):
                clean_name = str(col).replace("'", "").replace("[", "_").replace("]", "_").replace(" ", "_")
                record[clean_name] = row_data[i]
            return record
        else:
            # For other access patterns, return raw data
            return self.data[index]


class CSVStore(compass_model.Store):
    """A data store represented by a CSV file."""

    @staticmethod
    def plugin_name():
        return "CSV"

    @staticmethod
    def plugin_description():
        return "A plugin used to browse CSV files as array data."

    file_extensions = {'CSV File': ['*.csv'], 'Tab-separated': ['*.tsv'], 'Text Files': ['*.txt']}

    def __contains__(self, key):
        if key == '/':
            return True
        return False

    @property
    def url(self):
        return self._url

    @property
    def display_name(self):
        return op.basename(url2path(self._url))

    @property
    def root(self):
        return self['/']

    @property
    def valid(self):
        return self._valid

    @staticmethod
    def can_handle(url):
        if not url.startswith('file://'):
            log.debug("able to handle %s? no, not starting with file://" % url)
            return False

        path = url2path(url)
        if not op.exists(path):
            log.debug("able to handle %s? no, file does not exist" % url)
            return False

        # Check file extension
        _, ext = op.splitext(path.lower())
        if ext not in ['.csv', '.tsv', '.txt']:
            log.debug("able to handle %s? no, unsupported extension %s" % (url, ext))
            return False

        # Try to read a few lines to validate it's a CSV-like file
        try:
            df = pd.read_csv(path, nrows=5)
            if len(df.columns) == 0:
                log.debug("able to handle %s? no, no columns found" % url)
                return False
            log.debug("able to handle %s? yes" % url)
            return True
        except Exception as e:
            log.debug("able to handle %s? no, pandas read error: %s" % (url, str(e)))
            return False

    def __init__(self, url):
        if not self.can_handle(url):
            raise ValueError(url)
        self._url = url
        self._valid = True

    def close(self):
        self._valid = False

    def get_parent(self, key):
        return None

    def getFilePath(self):
        return url2path(self._url)


class CSVArray(compass_model.Array):
    """Represents CSV data as an array."""

    class_kind = "CSV Data"

    @staticmethod
    def can_handle(store, key):
        return key == '/'

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._data = None
        self._shape = None
        self._dtype = None
        self._columns = None

        # Load metadata without loading full data
        self._load_metadata()

    def _load_metadata(self):
        """Load just the metadata (shape, columns, dtype) without loading all data."""
        try:
            file_path = self._store.getFilePath()

            # Determine separator
            _, ext = op.splitext(file_path.lower())
            sep = '\t' if ext == '.tsv' else ','

            # Read just a sample to get column info and infer dtypes
            sample_df = pd.read_csv(file_path, sep=sep, nrows=100)

            # Get full row count more efficiently
            with open(file_path, 'r') as f:
                row_count = sum(1 for line in f) - 1  # -1 for header

            self._shape = (row_count, len(sample_df.columns))
            self._columns = sample_df.columns.tolist()

            # Determine overall dtype - use object for mixed types
            if sample_df.dtypes.nunique() == 1:
                # All columns have same type
                self._dtype = sample_df.dtypes.iloc[0]
            else:
                # Mixed types - use object
                self._dtype = np.dtype('object')

        except Exception as e:
            log.error("Error loading CSV metadata: %s" % str(e))
            self._shape = (0, 0)
            self._columns = []
            self._dtype = np.dtype('object')

    def _load_data(self):
        """Load the actual data when needed."""
        if self._data is None:
            try:
                file_path = self._store.getFilePath()

                # Determine separator
                _, ext = op.splitext(file_path.lower())
                sep = '\t' if ext == '.tsv' else ','

                # Load the data
                df = pd.read_csv(file_path, sep=sep)
                self._data = df.values

            except Exception as e:
                log.error("Error loading CSV data: %s" % str(e))
                # Create empty array as fallback
                self._data = np.empty(self._shape, dtype=self._dtype)

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return self._store.display_name

    @property
    def description(self):
        return 'CSV file "%s" with %d rows and %d columns' % (
            self.display_name, self._shape[0], self._shape[1])

    @property
    def shape(self):
        # For structured arrays, ArrayTable expects shape to be (num_records,)
        if self._columns:
            return (self._shape[0],)
        return self._shape

    @property
    def dtype(self):
        # Create a structured dtype so ArrayTable treats this as columnar data
        if self._columns:
            dtype_list = []
            for col in self._columns:
                # Clean column names for numpy dtype compatibility
                clean_name = str(col).replace("'", "").replace("[", "_").replace("]", "_").replace(" ", "_")
                dtype_list.append((clean_name, 'O'))
            return np.dtype(dtype_list)
        return self._dtype

    def __getitem__(self, args):
        self._load_data()

        # Handle structured array access for ArrayTable compatibility
        if isinstance(args, str):
            # Accessing by field name (column name)
            field_name = args
            for i, col in enumerate(self._columns):
                clean_name = str(col).replace("'", "").replace("[", "_").replace("]", "_").replace(" ", "_")
                if clean_name == field_name:
                    return self._data[:, i]
            raise KeyError(f"Field {field_name} not found")

        # Handle single integer index (row access) for structured arrays
        if isinstance(args, int):
            # Return a structured record-like object
            row_data = self._data[args, :]
            # Create a record-like object with named fields
            record = {}
            for i, col in enumerate(self._columns):
                clean_name = str(col).replace("'", "").replace("[", "_").replace("]", "_").replace(" ", "_")
                record[clean_name] = row_data[i]
            return record

        # Handle slice access
        if isinstance(args, slice):
            # Return multiple rows as a custom object that supports both array and structured access
            slice_data = self._data[args, :]
            return StructuredSlice(slice_data, self._columns)

        # Handle tuple slice access (for LRUTileCache)
        if isinstance(args, tuple) and len(args) == 1 and isinstance(args[0], slice):
            slice_data = self._data[args[0], :]
            return StructuredSlice(slice_data, self._columns)

        # Default behavior for other access patterns
        return self._data[args]


class CSVKeyValue(compass_model.KeyValue):
    """Key-value view of CSV metadata."""

    class_kind = "CSV Metadata"

    @staticmethod
    def can_handle(store, key):
        return key == '/'

    def __init__(self, store, key):
        self._store = store
        self._key = key

        # Get file metadata
        file_path = store.getFilePath()
        import os
        stat = os.stat(file_path)

        # Read sample to get column info
        try:
            _, ext = op.splitext(file_path.lower())
            sep = '\t' if ext == '.tsv' else ','
            sample_df = pd.read_csv(file_path, sep=sep, nrows=5)
            columns = sample_df.columns.tolist()
            dtypes = {str(col): str(dtype) for col, dtype in sample_df.dtypes.items()}
        except Exception:
            columns = []
            dtypes = {}

        self.data = {
            'File size (bytes)': stat.st_size,
            'Number of columns': len(columns),
            'Column names': ', '.join(columns) if columns else 'Unknown',
            'Column types': str(dtypes) if dtypes else 'Unknown',
            'Separator': 'Tab' if ext == '.tsv' else 'Comma',
        }

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return "CSV Info"

    @property
    def description(self):
        return "Metadata for CSV file"

    @property
    def keys(self):
        return list(self.data.keys())

    def __getitem__(self, key):
        return self.data[key]


# Register the components
CSVStore.push(CSVKeyValue)  # metadata
CSVStore.push(CSVArray)     # array data

compass_model.push(CSVStore)


#!/usr/bin/env python3
"""
Test for Context Interface API
Tests ContextBundle, ContextQuery, and ContextDestroy methods
"""

import sys
import os
import time
import tempfile
import socket

# Add current directory to path for module import
sys.path.insert(0, os.getcwd())

try:
    # Import the CEE API module (built by nanobind)
    import wrp_cee as cee
except ImportError as e:
    print(f"❌ Failed to import wrp_cee module: {e}")
    print("   Make sure WRP_CORE_ENABLE_PYTHON=ON and nanobind is installed")
    sys.exit(1)

# Import CTE module for Tag and runtime initialization
try:
    import wrp_cte_core_ext as cte
except ImportError as e:
    print(f"❌ Failed to import wrp_cte_core_ext module: {e}")
    print("   CTE Python bindings are required for runtime initialization")
    sys.exit(1)


def find_available_port(start_port=9129, end_port=9200):
    """Find an available port in the given range"""
    for port in range(start_port, end_port):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(('', port))
                return port
            except OSError:
                continue
    raise RuntimeError(f"No available ports in range {start_port}-{end_port}")


def initialize_runtime():
    """Initialize Chimaera runtime for testing"""
    try:
        import yaml
    except ImportError:
        print("❌ PyYAML not available - cannot run tests")
        return False

    temp_dir = tempfile.gettempdir()

    # Create hostfile
    hostfile = os.path.join(temp_dir, "wrp_context_test_hostfile")
    with open(hostfile, 'w') as f:
        f.write("localhost\n")

    # Find available port
    port = find_available_port()
    print(f"Using port: {port}")

    # Create storage directory
    storage_dir = os.path.join(temp_dir, "cte_context_test_storage")
    os.makedirs(storage_dir, exist_ok=True)

    # Generate config
    config = {
        'networking': {
            'protocol': 'zmq',
            'hostfile': hostfile,
            'port': port
        },
        'workers': {
            'num_workers': 4
        },
        'memory': {
            'main_segment_size': '1G',
            'client_data_segment_size': '512M',
            'runtime_data_segment_size': '512M'
        },
        'devices': [
            {
                'mount_point': storage_dir,
                'capacity': '1G'
            }
        ]
    }

    # Write config
    config_path = os.path.join(temp_dir, "wrp_context_test_conf.yaml")
    with open(config_path, 'w') as f:
        yaml.dump(config, f)

    # Set environment
    os.environ['CHI_SERVER_CONF'] = config_path
    os.environ['CHI_REPO_PATH'] = os.getcwd()

    try:
        # Initialize Chimaera (unified init)
        print("Initializing Chimaera...")
        if not cte.chimaera_init(cte.ChimaeraMode.kClient, True):
            print("❌ Chimaera init failed")
            return False
        time.sleep(0.5)

        # Initialize CTE
        print("Initializing CTE subsystem...")
        pool_query = cte.PoolQuery.Dynamic()
        if not cte.initialize_cte(config_path, pool_query):
            print("❌ CTE init failed")
            return False
        time.sleep(0.3)

        print("✅ Runtime initialized successfully\n")
        return True

    except Exception as e:
        print(f"❌ Runtime initialization failed: {e}")
        return False


def test_context_bundle():
    """Test ContextBundle method"""
    print("Test 1: ContextBundle")

    # Create test file
    temp_dir = tempfile.gettempdir()
    test_file = os.path.join(temp_dir, "test_bundle_data.bin")
    test_data = b"Hello from ContextBundle test!"

    with open(test_file, 'wb') as f:
        f.write(test_data)

    try:
        # Create ContextInterface instance
        ctx_interface = cee.ContextInterface()

        # Create AssimilationCtx
        ctx = cee.AssimilationCtx(
            src=f"file://{test_file}",
            dst="iowarp::test_bundle_tag",
            format="binary"
        )

        # Call ContextBundle
        result = ctx_interface.context_bundle([ctx])

        if result == 0:
            print(f"✅ ContextBundle succeeded")
            return True
        else:
            print(f"❌ ContextBundle failed with code: {result}")
            return False

    except Exception as e:
        print(f"❌ ContextBundle threw exception: {e}")
        import traceback
        traceback.print_exc()
        return False
    finally:
        # Cleanup
        if os.path.exists(test_file):
            os.remove(test_file)


def test_context_query():
    """Test ContextQuery method"""
    print("\nTest 2: ContextQuery")

    try:
        # First, put some test data using CTE Tag API
        tag = cte.Tag("test_query_tag")
        tag.put_blob("blob1", b"Test data 1")
        tag.put_blob("blob2", b"Test data 2")
        tag.put_blob("blob3", b"Test data 3")
        time.sleep(0.2)  # Allow operations to complete

        # Create ContextInterface instance
        ctx_interface = cee.ContextInterface()

        # Query for blobs
        results = ctx_interface.context_query("test_query_tag", ".*", 0)

        if results and isinstance(results, list):
            print(f"✅ ContextQuery returned {len(results)} results: {results}")
            return True
        else:
            print(f"❌ ContextQuery failed to return results")
            return False

    except Exception as e:
        print(f"❌ ContextQuery threw exception: {e}")
        import traceback
        traceback.print_exc()
        return False


def test_context_destroy():
    """Test ContextDestroy method"""
    print("\nTest 3: ContextDestroy")

    try:
        # Create a tag to destroy
        tag = cte.Tag("test_destroy_tag")
        tag.put_blob("temp_blob", b"Temporary data")
        time.sleep(0.2)

        # Create ContextInterface instance
        ctx_interface = cee.ContextInterface()

        # Destroy the tag
        result = ctx_interface.context_destroy(["test_destroy_tag"])

        if result == 0:
            print(f"✅ ContextDestroy succeeded")
            return True
        else:
            print(f"❌ ContextDestroy failed with code: {result}")
            return False

    except Exception as e:
        print(f"❌ ContextDestroy threw exception: {e}")
        import traceback
        traceback.print_exc()
        return False


def main():
    """Run all context interface tests"""
    print("=" * 70)
    print("Context Interface Test Suite (CEE API)")
    print("=" * 70)
    print()

    # Initialize runtime
    if not initialize_runtime():
        print("\n❌ Cannot run tests without runtime initialization")
        return 1

    # Run tests
    tests = [
        ("ContextBundle", test_context_bundle),
        ("ContextQuery", test_context_query),
        ("ContextDestroy", test_context_destroy),
    ]

    passed = 0
    total = len(tests)

    for test_name, test_func in tests:
        try:
            if test_func():
                passed += 1
        except Exception as e:
            print(f"❌ Test '{test_name}' threw exception: {e}")
            import traceback
            traceback.print_exc()

    # Summary
    print("\n" + "=" * 70)
    print(f"Test Results: {passed}/{total} tests passed")
    print("=" * 70)

    if passed == total:
        print("✅ All context interface tests passed!")
        return 0
    else:
        print(f"❌ {total - passed} test(s) failed")
        return 1


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
setup_collection_consent.py - Set up consent for HTTPS access to a Globus collection

This script handles getting the proper consent for accessing a guest collection
via HTTPS. Guest collections require collection-specific scopes.

Usage:
  ./setup_collection_consent.py <collection_id>
"""

import sys
from globus_cli.login_manager import LoginManager

def main():
    if len(sys.argv) < 2:
        print("Usage: setup_collection_consent.py <collection_id>")
        sys.exit(1)

    collection_id = sys.argv[1]

    print(f"Setting up HTTPS access consent for collection: {collection_id}")
    print("")

    # The scope needed for HTTPS access to this collection
    https_scope = f"https://auth.globus.org/scopes/{collection_id}/https"

    print(f"Required scope: {https_scope}")
    print("")
    print("This collection requires collection-specific consent.")
    print("You need to add this scope when logging in.")
    print("")
    print("Run the following command:")
    print(f"  globus session consent '{https_scope}'")
    print("")
    print("Or re-login with the required scopes:")
    print(f"  globus logout")
    print(f"  globus login")
    print("")
    print("Note: The Globus CLI may automatically request consent when needed.")

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from conan import ConanFile
from conan.tools.cmake import CMake, cmake_layout
from conan.tools.build import can_run
from conan.tools.files import save, load
import os


class TestPackageConan(ConanFile):
    settings = "os", "compiler", "build_type", "arch"
    generators = "CMakeToolchain", "CMakeDeps", "VirtualRunEnv"
    test_type = "explicit"

    def requirements(self):
        self.requires(self.tested_reference_str)

    def layout(self):
        cmake_layout(self)

    def generate(self):
        save(self, os.path.join(self.build_folder, "package_folder"),
             self.dependencies[self.tested_reference_str].package_folder)
        save(self, os.path.join(self.build_folder, "license"),
             self.dependencies[self.tested_reference_str].license)

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def test(self):
        if can_run(self):
            cmd = os.path.join(self.cpp.build.bindir, "test_package")
            self.run(cmd, env="conanrun")

            package_folder = load(self, os.path.join(self.build_folder, "package_folder"))
            license = load(self, os.path.join(self.build_folder, "license"))
            assert os.path.isfile(os.path.join(package_folder, "licenses", "LICENSE.txt"))
            assert license == 'BSL-1.0'



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
"""
Implementation of compass_model classes for BAG files.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from itertools import groupby
import sys
import os.path as op
import posixpath as pp
import h5py

# Try to import hydroffice.bag, fall back to h5py if not available
try:
    from hydroffice.bag import is_bag
    from hydroffice.bag import BAGFile
    from hydroffice.bag import BAGError
    HAS_HYDROFFICE_BAG = True
except ImportError:
    import h5py
    import numpy as np
    HAS_HYDROFFICE_BAG = False
    
    def is_bag(path):
        """Simple BAG file detection using h5py"""
        try:
            with h5py.File(path, 'r') as f:
                return 'BAG_root' in f
        except:
            return False
    
    class BAGError(Exception):
        pass
    
    class BAGFile:
        """Simple BAG file wrapper using h5py"""
        def __init__(self, path, mode='r'):
            self.filename = path
            self._file = h5py.File(path, mode)
            
        def close(self):
            self._file.close()
            
        def __contains__(self, key):
            return key in self._file
            
        def __getitem__(self, key):
            return self._file[key]
        
        def elevation(self, mask_nan=False):
            """Get elevation data"""
            data = self._file['BAG_root/elevation'][:]
            if mask_nan:
                data = np.ma.masked_invalid(data)
            return data
            
        def uncertainty(self, mask_nan=False):
            """Get uncertainty data"""
            data = self._file['BAG_root/uncertainty'][:]
            if mask_nan:
                data = np.ma.masked_invalid(data)
            return data
            
        def tracking_list(self):
            """Get tracking list data"""
            return self._file['BAG_root/tracking_list'][:]
            
        def metadata(self, as_string=True, as_pretty_xml=False):
            """Get metadata"""
            if as_string:
                data = self._file['BAG_root/metadata'][:]
                if isinstance(data, bytes):
                    return data.decode('utf-8')
                return str(data)
            return self._file['BAG_root/metadata'][:]
            
        def populate_metadata(self):
            """Simple metadata object"""
            class SimpleMeta:
                def geo_extent(self):
                    return (0, 1, 0, 1)  # Default extent
            return SimpleMeta()
            
        def validation_info(self):
            """Validation info placeholder"""
            return "Validation not available without hydroffice.bag package"

from hdf_compass import compass_model
from hdf_compass.utils import url2path

import logging
log = logging.getLogger(__name__)


def sort_key(name):
    """ Sorting key for names in an BAG group.

    We provide "natural" sort order; e.g. "7" comes before "12".
    """
    return [(int(''.join(g)) if k else ''.join(g)) for k, g in groupby(name, key=str.isdigit)]


class BAGStore(compass_model.Store):
    """ Data store implementation using a BAG file (closely mimicking HDF5Store).

    Keys are the full names of objects in the file.
    """
    @staticmethod
    def plugin_name():
        return "BAG"

    @staticmethod
    def plugin_description():
        return """A plugin used to browse Open Navigation Surface BAG files.
        It provides additional features that are not available with the general HDF5 plugin:
        - View of metadata information as XML (+ content validation).
        - Plot of elevation and uncertainty using their geographic extent/
        The plugin is developed and maintained by G.Masetti [gmasetti@ccom.unh.edu].
        """

    file_extensions = {'BAG File': ['*.bag']}

    def __contains__(self, key):
        return key in self.f

    @property
    def url(self):
        return self._url

    @property
    def display_name(self):
        return op.basename(self.f.filename)

    @property
    def root(self):
        return self['/']

    @property
    def valid(self):
        return bool(self.f)

    @staticmethod
    def can_handle(url):
        if not url.startswith('file://'):
            log.debug("able to handle %s? no, invalid url" % url)
            return False

        path = url2path(url)
        if not is_bag(path):
            log.debug("able to handle %s? no, not a BAG" % url)
            return False

        log.debug("able to handle %s? yes" % url)
        return True

    def __init__(self, url):
        if not self.can_handle(url):
            raise ValueError(url)
        self._url = url
        path = url2path(url)
        self.f = BAGFile(path, 'r')

    def close(self):
        self.f.close()

    def get_parent(self, key):
        # HDFCompass requires the parent of the root container be None
        if key == "" or key == "/":
            return None
        p_key = pp.dirname(key)
        if p_key == "":
            p_key = "/"
        return self[p_key]


class BAGGroup(compass_model.Container):
    """ Represents an BAG group (closely mimicking HDF5Group). """
    class_kind = "BAG Group"

    @staticmethod
    def can_handle(store, key):
        return (key in store) and (isinstance(store.f[key], h5py.Group))

    @property
    def _names(self):

        # Lazily build the list of names; this helps when browsing big files
        if self._xnames is None:

            self._xnames = list(self._group)

            # Natural sort is expensive
            if len(self._xnames) < 1000:
                self._xnames.sort(key=sort_key)

        return self._xnames

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._group = store.f[key]
        self._xnames = None

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        name = pp.basename(self.key)
        if name == "":
            name = '/'
        return name

    @property
    def display_title(self):
        return "%s %s" % (self.store.display_name, self.key)

    @property
    def description(self):
        return 'Group "%s" (%d members)' % (self.display_name, len(self))

    def __len__(self):
        return len(self._group)

    def __iter__(self):
        for name in self._names:
            yield self.store[pp.join(self.key, name)]

    def __getitem__(self, idx):
        name = self._names[idx]
        return self.store[pp.join(self.key, name)]


class BAGRoot(compass_model.Container):
    """ Represents the BAG root. """
    class_kind = "BAG Root"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root") and (key in store) and (isinstance(store.f[key], h5py.Group))

    @property
    def _names(self):

        # Lazily build the list of names; this helps when browsing big files
        if self._xnames is None:

            self._xnames = list(self._group)

            # Natural sort is expensive
            if len(self._xnames) < 1000:
                self._xnames.sort(key=sort_key)

        return self._xnames

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._group = store.f[key]
        self._xnames = None

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        name = pp.basename(self.key)
        return name

    @property
    def display_title(self):
        return "%s %s" % (self.store.display_name, self.key)

    @property
    def description(self):
        return 'Root Group "%s" (%d members)' % (self.display_name, len(self))

    def __len__(self):
        return len(self._group)

    def __iter__(self):
        for name in self._names:
            yield self.store[pp.join(self.key, name)]

    def __getitem__(self, idx):
        name = self._names[idx]
        return self.store[pp.join(self.key, name)]


class BAGDataset(compass_model.Array):
    """ Represents a BAG dataset (closely mimicking HDF5Dataset). """
    class_kind = "BAG Dataset"

    @staticmethod
    def can_handle(store, key):
        return key in store and isinstance(store.f[key], h5py.Dataset)

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f[key]

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]

    def is_plottable(self):
        if self.dtype.kind == 'S':
            log.debug("Not plottable since ASCII String (characters: %d)" % self.dtype.itemsize)
            return False
        if self.dtype.kind == 'U':
            log.debug("Not plottable since Unicode String (characters: %d)" % self.dtype.itemsize)
            return False
        return True


class BAGElevationArray(compass_model.Array):
    """ Represents a BAG elevation. """
    class_kind = "BAG Elevation [array]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/elevation") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.elevation(mask_nan=True)

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]


class BAGElevationGeoArray(compass_model.GeoArray):
    """ Represents a BAG elevation. """
    class_kind = "BAG Elevation [geo array]"

    @staticmethod
    def can_handle(store, key):
        # for GeoSurface we use cartopy that can be challenging to freeze on OSX to dependencies (i.e. geos)
        try:
            import cartopy.crs as ccrs
        except (ImportError, OSError):
            return False
        return (key == "/BAG_root/elevation") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.elevation(mask_nan=True)
        self._meta = store.f.populate_metadata()

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    @property
    def extent(self):
        """ Geographic extent as a tuple: (lon_min, lon_max, lon_min, lon_max) """
        return self._meta.geo_extent()

    def __getitem__(self, args):
        return self._dset[args]


class BAGElevation(compass_model.GeoSurface):
    """ Represents a BAG elevation. """
    class_kind = "BAG Elevation"

    @staticmethod
    def can_handle(store, key):
        # for GeoSurface we use cartopy that can be challenging to freeze on OSX to dependencies (i.e. geos)
        try:
            import cartopy.crs as ccrs
        except (ImportError, OSError):
            return False
        # for GeoSurface we are using a matplotlib function present after 1.5.x
        import matplotlib
        plt_maj, plt_min, _ = matplotlib.__version__.split('.')
        if (int(plt_maj) == 1) and (int(plt_min) < 5):
            return False
        return (key == "/BAG_root/elevation") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.elevation(mask_nan=True)
        self._meta = store.f.populate_metadata()

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    @property
    def extent(self):
        """ Geographic extent as a tuple: (x_min, x_max, y_min, y_max) """
        return self._meta.geo_extent()

    def __getitem__(self, args):
        return self._dset[args]


class BAGUncertaintyArray(compass_model.Array):
    """ Represents an uncertainty array. """
    class_kind = "BAG Uncertainty [array]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/uncertainty") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.uncertainty(mask_nan=True)

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]


class BAGTrackinList(compass_model.Array):
    """ Represents a BAG tracking list. """
    class_kind = "BAG Tracking List"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/tracking_list") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.tracking_list()

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]


class BAGMetadataRaw(compass_model.Array):
    """ Represents a raw BAG metadata. """
    class_kind = "BAG Metadata [raw]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/metadata") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.metadata(as_string=False, as_pretty_xml=False)

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]

    def is_plottable(self):
        if self.dtype.kind == 'S':
            log.debug("Not plottable since ASCII String (characters: %d)" % self.dtype.itemsize)
            return False
        if self.dtype.kind == 'U':
            log.debug("Not plottable since Unicode String (characters: %d)" % self.dtype.itemsize)
            return False
        return True


class BAGMetadataText(compass_model.Text):
    """ Represents a text BAG metadata. """

    class_kind = "BAG Metadata [text]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/metadata") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        try:
            self._dset = store.f.metadata(as_string=True, as_pretty_xml=True)
        except BAGError as e:
            log.warning("unable to retrieve metadata as xml")
            self._dset = ""

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def text(self):
        return self._dset


class BAGMetadataXml(compass_model.Xml):
    """ Represents a text BAG metadata. """

    class_kind = "BAG Metadata [xml]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/metadata") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def has_validation(self):
        """For BAG data there is a known validation mechanism based on XSD and Schematron"""
        return True

    def __init__(self, store, key):
        self._store = store
        self._key = key
        try:
            self._dset = store.f.metadata(as_string=True, as_pretty_xml=True)
        except BAGError as e:
            log.warning("unable to retrieve metadata as xml")
            self._dset = ""

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def text(self):
        return self._dset

    @property
    def validation(self):
        """ Collect a message string with the result of the validation """
        return self.store.f.validation_info()


class BAGUncertaintyArray(compass_model.Array):
    """ Represents an uncertainty array. """
    class_kind = "BAG Uncertainty [array]"

    @staticmethod
    def can_handle(store, key):
        return (key == "/BAG_root/uncertainty") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.uncertainty(mask_nan=True)

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    def __getitem__(self, args):
        return self._dset[args]


class BAGUncertainty(compass_model.GeoArray):
    """ Represents a BAG uncertainty. """
    class_kind = "BAG Uncertainty"

    @staticmethod
    def can_handle(store, key):
        # for GeoSurface we use cartopy that can be challenging to freeze on OSX to dependencies (i.e. geos)
        try:
            import cartopy.crs as ccrs
        except (ImportError, OSError):
            return False
        return (key == "/BAG_root/uncertainty") and (key in store) and (isinstance(store.f[key], h5py.Dataset))

    def __init__(self, store, key):
        self._store = store
        self._key = key
        self._dset = store.f.uncertainty(mask_nan=True)
        self._meta = store.f.populate_metadata()

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return pp.basename(self.key)

    @property
    def description(self):
        return 'Dataset "%s"' % (self.display_name,)

    @property
    def shape(self):
        return self._dset.shape

    @property
    def dtype(self):
        return self._dset.dtype

    @property
    def extent(self):
        """ Geographic extent as a tuple: (x_min, x_max, y_min, y_max) """
        return self._meta.geo_extent()

    def __getitem__(self, args):
        return self._dset[args]


class BAGKV(compass_model.KeyValue):
    """ A KeyValue node used for BAG attributes (closely mimicking HDF5KV). """
    class_kind = "BAG Attributes"

    @staticmethod
    def can_handle(store, key):
        return key in store.f

    def __init__(self, store, key):
        log.debug("init")
        self._store = store
        self._key = key
        self._obj = store.f[key]
        self._names = list(self._obj.attrs.keys())

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        n = pp.basename(self.key)
        return n if n != '' else '/'

    @property
    def description(self):
        return self.display_name

    @property
    def keys(self):
        return self._names[:]

    def __getitem__(self, name):
        return self._obj.attrs[name]


class BAGImage(compass_model.Image):
    """ BAG true-color images (closely mimicking HDF5Image). """
    class_kind = "BAG Truecolor Image"

    @staticmethod
    def can_handle(store, key):
        if key not in store:
            return False
        obj = store.f[key]
        if obj.attrs.get('CLASS') != 'IMAGE':
            return False
        if obj.attrs.get('IMAGE_SUBCLASS') != 'IMAGE_TRUECOLOR':
            return False
        if obj.attrs.get('INTERLACE_MODE') != 'INTERLACE_PIXEL':
            return False
        return True

    def __init__(self, store, key):
        log.debug("init")
        self._store = store
        self._key = key
        self._obj = store.f[key]

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        n = pp.basename(self.key)
        return n if n != '' else '/'

    @property
    def description(self):
        return self.display_name

    @property
    def width(self):
        return self._obj.shape[1]

    @property
    def height(self):
        return self._obj.shape[0]

    @property
    def palette(self):
        return None

    @property
    def data(self):
        return self._obj[:]


# Register handlers
BAGStore.push(BAGKV)
BAGStore.push(BAGDataset)
BAGStore.push(BAGElevationArray)
BAGStore.push(BAGElevationGeoArray)
BAGStore.push(BAGElevation)
BAGStore.push(BAGUncertaintyArray)
BAGStore.push(BAGUncertainty)
BAGStore.push(BAGTrackinList)
BAGStore.push(BAGMetadataRaw)
BAGStore.push(BAGMetadataText)
BAGStore.push(BAGMetadataXml)
BAGStore.push(BAGGroup)
BAGStore.push(BAGRoot)
BAGStore.push(BAGImage)

compass_model.push(BAGStore)



import test_inter_module_1_ext as t1
import test_inter_module_2_ext as t2
import test_classes_ext as t3
import pytest
from common import xfail_on_pypy_darwin

try:
    from concurrent import interpreters  # Added in Python 3.14
    def needs_interpreters(x):
        return x
except:
    needs_interpreters = pytest.mark.skip(reason="interpreters required")


@xfail_on_pypy_darwin
def test01_inter_module():
    s = t1.create_shared()
    assert t2.check_shared(s, 123)
    t2.increment_shared(s)
    assert t2.check_shared(s, 124)
    with pytest.raises(TypeError) as excinfo:
        assert t3.check_shared(s)
    assert 'incompatible function arguments' in str(excinfo.value)


@xfail_on_pypy_darwin
def test02_reload_module():
    s1 = t1.create_shared()
    s2 = t1.create_shared()
    assert s2 is not s1
    assert type(s2) is type(s1)
    t2.increment_shared(s2)
    import importlib
    new_t1 = importlib.reload(t1)
    assert new_t1 is t1
    s3 = new_t1.create_shared()
    assert type(s3) is type(s1)
    new_t2 = importlib.reload(t2)
    assert new_t2 is t2
    s4 = new_t1.create_shared()
    assert type(s4) is type(s1)
    assert new_t2.check_shared(s2, 124)


@xfail_on_pypy_darwin
def test03_reimport_module():
    s1 = t1.create_shared()
    s2 = t1.create_shared()
    t2.increment_shared(s2)
    import sys
    del sys.modules['test_inter_module_1_ext']
    import test_inter_module_1_ext as new_t1
    assert new_t1 is not t1
    s3 = new_t1.create_shared()
    assert type(s3) is type(s1)
    del sys.modules['test_inter_module_2_ext']
    with pytest.warns(RuntimeWarning, match="'Shared' was already registered"):
        import test_inter_module_2_ext as new_t2
    assert new_t2 is not t2
    s4 = new_t1.create_shared()
    assert type(s4) is type(s1)
    assert new_t2.check_shared(s2, 124)


def run():
    import sys
    if 'tests' not in sys.path[0]:
        import os
        builddir = sys.path[0]
        sys.path.insert(0, os.path.join(builddir, 'tests', 'Release'))
        sys.path.insert(0, os.path.join(builddir, 'tests', 'Debug'))
        sys.path.insert(0, os.path.join(builddir, 'tests'))
    import test_inter_module_1_ext as new_t1
    import test_inter_module_2_ext as new_t2
    success = True
    s = new_t1.create_shared()
    success &= new_t2.check_shared(s, 123)
    new_t2.increment_shared(s)
    success &= new_t2.check_shared(s, 124)
    return success

@needs_interpreters
def test04_subinterpreters():
    assert run()
    interp = interpreters.create()
    with pytest.raises(interpreters.ExecutionFailed) as excinfo:
        assert interp.call(run)
    assert 'does not support loading in subinterpreters' in str(excinfo.value)
    interp.close()



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
"""
Implements a viewer frame for compass_model.Array.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx
import wx.grid
from wx.lib.newevent import NewCommandEvent

import numpy as np

import os
import logging
import numpy

log = logging.getLogger(__name__)

from ..frame import NodeFrame
from .plot import LinePlotFrame, LineXYPlotFrame, ContourPlotFrame

class ArrayPanel(wx.Panel):
    """ Simple array panel with grid display """
    def __init__(self, parent, node):
        super(ArrayPanel, self).__init__(parent)
        self.node = node

        # Create a simple grid to show the array data
        self.grid = wx.grid.Grid(self)

        # Set up grid based on array shape
        try:
            if len(node.shape) == 0:  # scalar
                self.grid.CreateGrid(1, 1)
                try:
                    value = node[()]
                    self.grid.SetCellValue(0, 0, str(value))
                except Exception as e:
                    self.grid.SetCellValue(0, 0, f"<Error: {e}>")
            elif len(node.shape) == 1:  # 1D array
                rows = min(node.shape[0], 1000)  # Limit display for performance
                self.grid.CreateGrid(rows, 1)
                try:
                    data = node[:rows]
                    for i in range(rows):
                        try:
                            value = data[i]
                            # Handle numpy types and bytes
                            if hasattr(value, 'item'):
                                value = value.item()
                            if isinstance(value, bytes):
                                value = value.decode('utf-8', errors='replace')
                            self.grid.SetCellValue(i, 0, str(value))
                        except Exception as e:
                            self.grid.SetCellValue(i, 0, f"<Error: {e}>")
                except Exception as e:
                    # If we can't read the data at all, show an error message
                    self.grid.SetCellValue(0, 0, f"<Error reading data: {e}>")
            else:  # 2D or higher
                rows = min(node.shape[0], 100)
                cols = min(node.shape[1], 100) if len(node.shape) > 1 else 1
                self.grid.CreateGrid(rows, cols)
                if len(node.shape) == 2:
                    try:
                        data = node[:rows, :cols]
                        for i in range(rows):
                            for j in range(cols):
                                try:
                                    value = data[i, j]
                                    # Handle numpy types and bytes
                                    if hasattr(value, 'item'):
                                        value = value.item()
                                    if isinstance(value, bytes):
                                        value = value.decode('utf-8', errors='replace')
                                    self.grid.SetCellValue(i, j, str(value))
                                except Exception as e:
                                    self.grid.SetCellValue(i, j, f"<Error: {e}>")
                    except Exception as e:
                        # If we can't read the data at all, show an error message
                        self.grid.SetCellValue(0, 0, f"<Error reading data: {e}>")
        except Exception as e:
            # Fallback: create a minimal grid with error message
            if not hasattr(self, 'grid') or self.grid.GetNumberRows() == 0:
                self.grid.CreateGrid(1, 1)
            self.grid.SetCellValue(0, 0, f"<Error initializing grid: {e}>")

        # Layout
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(self.grid, 1, wx.EXPAND)
        self.SetSizer(sizer)
        
    def copy(self):
        """ Copy selected data to clipboard """
        pass  # TODO: Implement
        
    def export(self, path):
        """ Export data to CSV file """
        pass  # TODO: Implement
        
    def plot_data(self):
        """ Plot the data """
        try:
            from .plot import LinePlotFrame, ContourPlotFrame
            import numpy as np

            # Get the array frame to access the node data
            array_frame = self.GetParent()
            node = array_frame.node

            # Check if columns are selected in the grid
            selected_cols = self.grid.GetSelectedCols()
            selected_rows = self.grid.GetSelectedRows()

            # Get the raw data from the node
            data = node[:]

            # Handle column selection for plotting
            if len(selected_cols) > 0 and len(data.shape) >= 2:
                # Plot selected columns as separate line plots
                selected_data = []
                names = []
                for col in selected_cols:
                    if col < data.shape[1]:
                        selected_data.append(data[:, col])
                        names.append(f"Column {col}")

                if selected_data:
                    plot_frame = LinePlotFrame(selected_data, names, title=f"Plot (selected columns): {node.display_name}")
                    plot_frame.Show()
                    return

            # Handle row selection for plotting
            elif len(selected_rows) > 0 and len(data.shape) >= 1:
                # Plot selected rows as separate line plots
                selected_data = []
                names = []
                for row in selected_rows:
                    if row < data.shape[0]:
                        if len(data.shape) == 1:
                            # For 1D data, just plot the single value (not very useful)
                            selected_data.append([data[row]])
                        else:
                            # For 2D+ data, plot the row as a line
                            selected_data.append(data[row, :])
                        names.append(f"Row {row}")

                if selected_data:
                    plot_frame = LinePlotFrame(selected_data, names, title=f"Plot (selected rows): {node.display_name}")
                    plot_frame.Show()
                    return

            # No selection - plot all data using original logic
            # Determine plot type based on dimensionality
            if len(data.shape) == 0:
                wx.MessageBox("Cannot plot scalar data", "Plot Error", wx.OK | wx.ICON_ERROR)
                return
            elif len(data.shape) == 1:
                # 1D data - line plot
                plot_frame = LinePlotFrame([data], ["Data"], title=f"Plot: {node.display_name}")
            elif len(data.shape) == 2:
                # 2D data - contour plot
                plot_frame = ContourPlotFrame(data, ["Data"], title=f"Contour: {node.display_name}")
            else:
                # Multi-dimensional - plot first 2D slice
                if data.shape[0] == 1:
                    data_2d = data[0]
                else:
                    data_2d = data[0]
                plot_frame = ContourPlotFrame(data_2d, ["Data"], title=f"Contour (slice): {node.display_name}")

            plot_frame.Show()

        except Exception as e:
            wx.MessageBox(f"Error creating plot: {e}", "Plot Error", wx.OK | wx.ICON_ERROR)
        
    def plot_xy(self):
        """ Plot XY data """
        try:
            from .plot import LineXYPlotFrame
            import numpy as np

            # Get the array frame to access the node data
            array_frame = self.GetParent()
            node = array_frame.node

            # Check if columns are selected in the grid
            selected_cols = self.grid.GetSelectedCols()

            # Get the raw data from the node
            data = node[:]

            # Handle column selection for XY plotting
            if len(selected_cols) >= 2 and len(data.shape) >= 2:
                # Use first selected column as X and subsequent columns as Y series
                selected_cols = sorted(selected_cols)
                x_col = selected_cols[0]
                x_data = data[:, x_col]
                plot_data = [x_data]
                names = [f"Column {x_col}"]

                # Add all other selected columns as Y series
                for y_col in selected_cols[1:]:
                    if y_col < data.shape[1]:
                        plot_data.append(data[:, y_col])
                        names.append(f"Column {y_col}")

                plot_frame = LineXYPlotFrame(plot_data, names, title=f"XY Plot (selected columns): {node.display_name}")
                plot_frame.Show()
                return

            elif len(selected_cols) == 1 and len(data.shape) >= 2:
                # Only one column selected - use row indices as X
                col = selected_cols[0]
                if col < data.shape[1]:
                    x_data = np.arange(data.shape[0])
                    y_data = data[:, col]
                    plot_data = [x_data, y_data]
                    names = ["Row Index", f"Column {col}"]

                    plot_frame = LineXYPlotFrame(plot_data, names, title=f"XY Plot (Column {col}): {node.display_name}")
                    plot_frame.Show()
                    return

            # No selection or insufficient selection - use original logic
            # For XY plot, we need at least 2D data or 1D data that can be split
            if len(data.shape) == 0:
                wx.MessageBox("Cannot create XY plot from scalar data", "Plot Error", wx.OK | wx.ICON_ERROR)
                return
            elif len(data.shape) == 1:
                # 1D data - create X as indices and Y as data values
                x_data = np.arange(len(data))
                y_data = data
                plot_data = [x_data, y_data]
                names = ["Index", "Value"]
            elif len(data.shape) == 2:
                # 2D data - use first column as X and second as Y
                if data.shape[1] >= 2:
                    x_data = data[:, 0]
                    y_data = data[:, 1]
                    plot_data = [x_data, y_data]
                    names = ["Column 0", "Column 1"]
                else:
                    # Only one column - use indices as X
                    x_data = np.arange(data.shape[0])
                    y_data = data[:, 0]
                    plot_data = [x_data, y_data]
                    names = ["Row Index", "Value"]
            else:
                wx.MessageBox("XY plotting not supported for data with more than 2 dimensions", "Plot Error", wx.OK | wx.ICON_ERROR)
                return

            # Create XY plot
            plot_frame = LineXYPlotFrame(plot_data, names, title=f"XY Plot: {node.display_name}")
            plot_frame.Show()

        except Exception as e:
            wx.MessageBox(f"Error creating XY plot: {e}", "Plot Error", wx.OK | wx.ICON_ERROR)


# Indicates that the slicing selection may have changed.
# These events are emitted by the SlicerPanel.
ArraySlicedEvent, EVT_ARRAY_SLICED = NewCommandEvent()
ArraySelectionEvent, EVT_ARRAY_SELECTED = NewCommandEvent()

# Menu and button IDs
ID_VIS_MENU_PLOT = wx.NewId()
ID_VIS_MENU_PLOTXY = wx.NewId()
ID_VIS_MENU_COPY = wx.NewId()
ID_VIS_MENU_EXPORT = wx.NewId()

def gen_csv(data, delimiters):
    """ converts any N-dimensional array to a CSV-string """
    if(type(data) == numpy.ndarray or type(data) == list):
        return delimiters[0].join(map(lambda x: gen_csv(x,delimiters[1:]), data))
    else:
        return str(data)

class ArrayFrame(NodeFrame):
    """
    Top-level frame displaying objects of type compass_model.Array.

    From top to bottom, has:

    1. Toolbar (see ArrayFrame.init_toolbar)
    2. SlicerPanel, with controls for changing what's displayed.
    3. An ArrayGrid, which displays the data in a spreadsheet-like view.
    """

    last_open_csv = os.getcwd()
    csv_delimiters_copy = ['\n', '\t']
    csv_delimiters_export = ['\n', ',']

    def __init__(self, node, pos=None):
        """ Create a new array frame """
        super(ArrayFrame, self).__init__(node, size=(800, 400), title=node.display_name, pos=pos)

        # Create array-specific toolbar (replace any existing toolbar)
        existing_toolbar = self.GetToolBar()
        if existing_toolbar:
            existing_toolbar.Destroy()

        self.toolbar = self.CreateToolBar()
        self.toolbar.AddTool(wx.ID_COPY, "Copy", wx.ArtProvider.GetBitmap(wx.ART_COPY))
        self.toolbar.AddTool(wx.ID_SAVEAS, "Export", wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE))
        # Use specific plotting icons
        plot_icon_path = os.path.join(os.path.dirname(__file__), "..", "icons", "viz_plot_32.png")
        plot_xy_icon_path = os.path.join(os.path.dirname(__file__), "..", "icons", "viz_plot_xy_24.png")
        self.toolbar.AddTool(wx.ID_ZOOM_IN, "Plot Data", wx.Bitmap(plot_icon_path, wx.BITMAP_TYPE_PNG))
        self.toolbar.AddTool(wx.ID_ZOOM_OUT, "Plot XY", wx.Bitmap(plot_xy_icon_path, wx.BITMAP_TYPE_PNG))
        self.toolbar.Realize()

        # Initialize dimension selection for multidimensional arrays first
        if len(node.shape) > 2:
            self.rowSpin = wx.SpinCtrl(self, max=len(node.shape)-1, value="0", min=0)
            self.colSpin = wx.SpinCtrl(self, max=len(node.shape)-1, value="1", min=0)
            self.rowSpin.Bind(wx.EVT_SPINCTRL, self.on_dimSpin)
            self.colSpin.Bind(wx.EVT_SPINCTRL, self.on_dimSpin)
        else:
            # For 1D and 2D arrays, set default row/col indices
            self.rowSpin = None
            self.colSpin = None

        # Create the proper array view components first (before layout)
        # Create slicer panel for dimension control (parent must be ArrayFrame for indices property)
        has_names = node.dtype is not None and node.dtype.names is not None
        self.slicer = SlicerPanel(self, node.shape, has_names)

        # Layout components
        main_panel = wx.Panel(self)
        sizer = wx.BoxSizer(wx.VERTICAL)

        # Create grid for data display - need to pass ArrayFrame reference for table
        self.grid = ArrayGrid(main_panel, node, self.slicer, array_frame=self)

        # Add slicer if needed (reparent it to main_panel for proper layout)
        if len(node.shape) > (1 if has_names else 2):
            self.slicer.Reparent(main_panel)
            sizer.Add(self.slicer, 0, wx.EXPAND | wx.ALL, 5)

        # Add grid
        sizer.Add(self.grid, 1, wx.EXPAND | wx.ALL, 5)

        main_panel.SetSizer(sizer)
        self.view = main_panel

        # Bind events
        self.Bind(wx.EVT_TOOL, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_TOOL, self.on_export, id=wx.ID_SAVEAS)
        self.Bind(wx.EVT_TOOL, self.on_plot_data, id=wx.ID_ZOOM_IN)
        self.Bind(wx.EVT_TOOL, self.on_plot_xy, id=wx.ID_ZOOM_OUT)

        # Bind array events
        self.Bind(EVT_ARRAY_SLICED, self.on_sliced)
        self.Bind(EVT_ARRAY_SELECTED, self.on_selected)

        # Enable slicer controls after a short delay (wxPython Mac bug workaround)
        if hasattr(self.slicer, 'enable_spinctrls'):
            self.timer = wx.Timer(self)
            self.timer.Start(100, True)
            self.Bind(wx.EVT_TIMER, self.on_workaround_timer)

    def on_copy(self, evt):
        """ Handle copy toolbar button """
        # Get selected data and copy to clipboard
        data, names, line = self.get_selected_data()
        if data is not None:
            csv_data = gen_csv(data, self.csv_delimiters_copy)
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(csv_data))
                wx.TheClipboard.Close()

    def on_export(self, evt):
        """ Handle export toolbar button """
        with wx.FileDialog(self, "Export array", wildcard="CSV files (*.csv)|*.csv",
                          style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as dlg:
            if dlg.ShowModal() == wx.ID_OK:
                path = dlg.GetPath()
                data, names, line = self.get_selected_data()
                if data is not None:
                    csv_data = gen_csv(data, self.csv_delimiters_export)
                    with open(path, 'w') as f:
                        f.write(csv_data)

    def on_plot_data(self, evt):
        """ Handle plot data toolbar button """
        data, names, line = self.get_selected_data()
        if data is not None:
            if line:
                from .plot import LinePlotFrame
                plot_frame = LinePlotFrame(data, names, title=f"Plot: {self.node.display_name}")
                plot_frame.Show()
            else:
                from .plot import ContourPlotFrame
                plot_frame = ContourPlotFrame(data, names, title=f"Contour: {self.node.display_name}")
                plot_frame.Show()

    def on_plot_xy(self, evt):
        """ Handle plot XY toolbar button """
        data, names, line = self.get_selected_data()
        if data is not None and line:
            from .plot import LineXYPlotFrame
            # For XY plot, we need at least 2 data series
            if len(data) >= 2:
                plot_frame = LineXYPlotFrame(data, names, title=f"XY Plot: {self.node.display_name}")
                plot_frame.Show()
            else:
                wx.MessageBox("XY plotting requires at least 2 data series", "Plot Error", wx.OK | wx.ICON_ERROR)

    def on_selected(self, evt):
        """ User has chosen to display a different part of the dataset. """
        idx = 0
        for x in self.indices:
            self.slicer.set_spin_max(idx, self.node.shape[x]-1)
            idx = idx + 1
        
        self.grid.ResetView()

    def get_selected_data(self):
        """
        function to get the selected data in an array
        returns (data, names, line)
            data: array of sliced data
            names: name array for plots
            line: bool-value, True if 1D-Line, False if 2D
        """
        cols = self.grid.GetSelectedCols()
        rows = self.grid.GetSelectedRows()
        rank = len(self.node.shape)
        
        # Scalar data can't be line-plotted.
        if rank == 0:
            return None, None, True

        # Get data currently in the grid
        has_field_names = self.node.dtype is not None and self.node.dtype.names is not None
        if rank > 1 and not has_field_names:
            args = []
            for x in range(rank):
                if x == self.row:
                    args.append(slice(None, None, None))
                elif x == self.col:
                    args.append(slice(None, None, None))
                else:
                    idx = 0
                    for y in self.indices:
                        if y == x:
                            args.append(self.slicer.indices[idx])
                            break
                        idx = idx + 1
            data = self.node[tuple(args)]
            if self.row > self.col:
                data = np.transpose(data)
        else:
            # For structured arrays, use slice access to maintain field name support
            if has_field_names and self.slicer.indices == ():
                data = self.node[:]
            else:
                data = self.node[self.slicer.indices]

        # Columns in the view are selected
        if len(cols) != 0:
            # The data is compound
            if has_field_names:
                names = [self.grid.GetColLabelValue(x) for x in cols]
                data = [data[n] for n in names]
                return data, names, True

            # Plot multiple columns independently
            else:
                if rank > 1:
                    data = [data[(slice(None, None, None),c)] for c in cols]

                names = ["Col %d" % c for c in cols] if len(data) > 1 else None
                return data, names, True

        # Rows in view are selected
        elif len(rows) != 0:
            
            data = [data[(r,)] for r in rows]
            names = ["Row %d" % r for r in rows] if len(data) > 1 else None
            return data, names, True
        
        # No row or column selection.  Plot everything
        else:
            # The data is compound
            if has_field_names:
                names = [self.grid.GetColLabelValue(x) for x in range(self.grid.GetNumberCols())]
                data = [data[n] for n in names]
                return data, names, True

            # Plot 1D
            elif rank == 1:
                return [data], [], True

            # Plot 2D
            else:
                return data, [], False

    def on_sliced(self, evt):
        """ User has chosen to display a different part of the dataset. """
        self.grid.Refresh()

    def on_workaround_timer(self, evt):
        """ See slicer.enable_spinctrls docs """
        self.timer.Destroy()
        self.slicer.enable_spinctrls()

    @property
    def indices(self):
        """ A tuple of integer indices appropriate for dim selection.

        """
        l = []
        for x in range(len(self.node.shape)):
            if x == self.row or x == self.col:
                continue    
            l.append(x)
        return tuple(l)
        
    @property
    def row(self):
        """ The dimension selected for the row
        """
        if self.rowSpin is not None:
            return self.rowSpin.GetValue()
        return 0  # Default to first dimension

    @property
    def col(self):
        """ The dimension selected for the column
        """
        if self.colSpin is not None:
            return self.colSpin.GetValue()
        return 1 if len(self.node.shape) > 1 else 0  # Default to second dimension or first if 1D
        
        
    def on_dimSpin(self, evt):
        """ Dimension Spinbox value changed; notify parent to refresh the grid. """
        if self.rowSpin is None or self.colSpin is None:
            return

        pos = evt.GetPosition()
        otherSpinner = self.rowSpin

        if evt.GetEventObject() == self.rowSpin:
            otherSpinner = self.colSpin

        if pos == otherSpinner.GetValue():
            if (pos > 0):
                pos = pos - 1
            else:
                pos = pos + 1
            otherSpinner.SetValue(pos)

        wx.PostEvent(self, ArraySelectionEvent(self.GetId()))

class SlicerPanel(wx.Panel):
    """
    Holds controls for data access.

    Consult the "indices" property, which returns a tuple of indices that
    prefix the array.  This will be RANK-2 elements long, unless hasfields
    is true, in which case it will be RANK-1 elements long.
    """

    @property
    def indices(self):
        """ A tuple of integer indices appropriate for slicing.

        Will be RANK-2 elements long, RANK-1 if compound data is in use
        (hasfields == True).
        """
        return tuple([x.GetValue() for x in self.spincontrols])

    def __init__(self, parent, shape, hasfields):
        """ Create a new slicer panel.

        parent:     The wxPython parent window
        shape:      Shape of the data to visualize
        hasfields:  If True, the data is compound and the grid can only
                    display one axis.  So, we should display an extra spinbox.
        """
        wx.Panel.__init__(self, parent)
        self.parent = parent
        self.shape = shape
        self.hasfields = hasfields
        self.spincontrols = []

        # Rank of the underlying array
        rank = len(shape)

        # Rank displayable in the grid.  If fields are present, they occupy
        # the columns, so the data displayed is actually 1-D.
        visible_rank = 1 if hasfields else 2

        sizer = wx.BoxSizer(wx.HORIZONTAL)  # Will arrange the SpinCtrls
                
        if rank > visible_rank:
            infotext = wx.StaticText(self, wx.ID_ANY, "Array Indexing: ")
            sizer.Add(infotext, 0, flag=wx.EXPAND | wx.ALL, border=10)

            for idx in range(rank - visible_rank):
                maxVal = shape[idx] - 1
                if not hasfields:
                    maxVal = shape[self.parent.indices[idx]] - 1
                sc = wx.SpinCtrl(self, max=maxVal, value="0", min=0)
                sizer.Add(sc, 0, flag=wx.EXPAND | wx.ALL, border=10)
                sc.Disable()
                self.spincontrols.append(sc)

        self.SetSizer(sizer)

        self.Bind(wx.EVT_SPINCTRL, self.on_spin)

    def enable_spinctrls(self):
        """ Unlock the spin controls.

        Because of a bug in wxPython on Mac, by default the first spin control
        has bizarre contents (and control focus) when the panel starts up.
        Call this after a short delay (e.g. 100 ms) to enable indexing.
        """
        for sc in self.spincontrols:
            sc.Enable()

    def set_spin_max(self, idx, max):
        self.spincontrols[idx].SetRange(0, max)
        
    def on_spin(self, evt):
        """ Spinbox value changed; notify parent to refresh the grid. """
        wx.PostEvent(self, ArraySlicedEvent(self.GetId()))


class ArrayGrid(wx.grid.Grid):
    """
    Grid class to display the Array.

    Cell contents and appearance are handled by the table model in ArrayTable.
    """

    def __init__(self, parent, node, slicer, array_frame=None):
        wx.grid.Grid.__init__(self, parent)
        # Use array_frame if provided, otherwise use parent (for backward compatibility)
        table_parent = array_frame if array_frame is not None else parent
        table = ArrayTable(table_parent)
        self.SetTable(table, True)

        # Set selection mode - try different wxPython versions
        has_field_names = node.dtype is not None and node.dtype.names is not None
        try:
            # Modern wxPython (4.x)
            selmode = wx.grid.Grid.GridSelectionModes.GridSelectColumns
            if not has_field_names and len(node.shape) > 1:
                selmode = wx.grid.Grid.GridSelectionModes.GridSelectRowsAndColumns
        except AttributeError:
            try:
                # Older wxPython versions
                selmode = wx.grid.Grid.wxGridSelectColumns
                if not has_field_names and len(node.shape) > 1:
                    selmode |= wx.grid.Grid.wxGridSelectRows
            except AttributeError:
                try:
                    # Try without wx prefix
                    selmode = wx.grid.GridSelectColumns
                    if not has_field_names and len(node.shape) > 1:
                        selmode |= wx.grid.GridSelectRows
                except AttributeError:
                    # Last resort - use numeric constants
                    selmode = 2  # GridSelectColumns
                    if not has_field_names and len(node.shape) > 1:
                        selmode |= 1  # GridSelectRows

        try:
            self.SetSelectionMode(selmode)
        except Exception:
            # If setting selection mode fails, just continue without it
            pass
           
    def ResetView(self):
            """Trim/extend the grid if needed"""
            rowChange = self.GetTable().GetRowsCount() - self.GetNumberRows()
            colChange = self.GetTable().GetColsCount() - self.GetNumberCols()
            if rowChange != 0 or colChange != 0:
                self.ClearGrid()
                locker = wx.grid.GridUpdateLocker(self)
                if rowChange > 0:
                    msg = wx.grid.GridTableMessage(
                        self.GetTable(),
                        wx.grid.GRIDTABLE_NOTIFY_ROWS_APPENDED,
                        rowChange
                    )
                    self.ProcessTableMessage(msg)
                elif rowChange < 0:
                    msg = wx.grid.GridTableMessage(
                        self.GetTable(),
                        wx.grid.GRIDTABLE_NOTIFY_ROWS_DELETED,
                        0,
                        -rowChange
                    )
                    self.ProcessTableMessage(msg)
                    
                if colChange > 0:
                    msg = wx.grid.GridTableMessage(
                        self.GetTable(),
                        wx.grid.GRIDTABLE_NOTIFY_COLS_APPENDED,
                        colChange
                    )
                    self.ProcessTableMessage(msg)
                elif colChange < 0:
                    msg = wx.grid.GridTableMessage(
                        self.GetTable(),
                        wx.grid.GRIDTABLE_NOTIFY_COLS_DELETED,
                        0,
                        -colChange
                    )
                    self.ProcessTableMessage(msg)

            # The scroll bars aren't resized (at least on windows)
            # Jiggling the size of the window rescales the scrollbars
            # h,w = self.GetSize()
            # self.SetSize((h+1, w))
            # self.SetSize((h, w))
            self.ForceRefresh()


class LRUTileCache(object):
    """
        Simple tile-based LRU cache which goes between the Grid and
        the Array object.  Caches tiles along the last 1 or 2 dimensions
        of a dataset.

        Access is via __getitem__.  Because this class exists specifically
        to support point-based callbacks for the Grid, arguments may
        only be indices, not slices.
    """

    TILESIZE = 100  # Tiles will have shape (100,) or (100, 100)
    MAXTILES = 50  # Max number of tiles to retain in the cache

    def __init__(self, arr):
        """ *arr* is anything implementing compass_model.Array """
        import collections
        self.cache = collections.OrderedDict()
        self.arr = arr

    def __getitem__(self, args):
        """ Restricted to an index or tuple of indices. """

        if not isinstance(args, tuple):
            args = (args,)

        # Split off the last 1 or 2 dimensions
        coarse_position, fine_position = args[0:-2], args[-2:]

        def clip(x):
            """ Round down to nearest TILESIZE; takes e.g. 181 -> 100 """
            return (x // self.TILESIZE) * self.TILESIZE

        # Tuple with index of tile corner
        tile_key = coarse_position + tuple(clip(x) for x in fine_position)

        # Slice which will be applied to dataset to retrieve tile
        tile_slice = coarse_position + tuple(slice(clip(x), clip(x) + self.TILESIZE) for x in fine_position)

        # Index applied to tile to retrieve the desired data point
        tile_data_index = tuple(x % self.TILESIZE for x in fine_position)

        # Case 1: Add tile to cache, ejecting oldest tile if needed
        if not tile_key in self.cache:

            if len(self.cache) >= self.MAXTILES:
                self.cache.popitem(last=False)

            # Use the array's __getitem__ method to ensure proper structured data handling
            # This is important for CSV/Parquet files that return structured records
            if hasattr(self.arr, '__getitem__'):
                tile = self.arr.__getitem__(tile_slice)
            else:
                tile = self.arr[tile_slice]
            self.cache[tile_key] = tile

        # Case 2: Mark the tile as recently accessed
        else:
            tile = self.cache.pop(tile_key)
            self.cache[tile_key] = tile

        return tile[tile_data_index]


class ArrayTable(wx.grid.PyGridTableBase):
    """
    "Table" class which provides data and metadata for the grid to display.

    The methods defined here define the contents of the table, as well as
    the number of rows, columns and their values.
    """

    def __init__(self, parent):
        """ Create a new Table instance for use with a grid control.

        node:     An compass_model.Array implementation instance.
        slicer:   An instance of SlicerPanel, so we can see what indices the
                  user has requested.
        """
        wx.grid.PyGridTableBase.__init__(self)

        self.node = parent.node
        self.selecter = parent
        self.slicer = parent.slicer

        self.rank = len(self.node.shape)
        self.names = self.node.dtype.names if self.node.dtype is not None else None

        self.cache = LRUTileCache(self.node)

    def GetNumberRows(self):
        """ Callback for number of rows displayed by the grid control """
        if self.rank == 0:
            return 1
        elif self.rank == 1:
            return self.node.shape[0]
        elif self.names is not None:
            return self.node.shape[-1]
        return self.node.shape[self.selecter.row]

    def GetNumberCols(self):
        """ Callback for number of columns displayed by the grid control.

        Note that if compound data is in use, columns display the field names.
        """
        if self.names is not None:
            return len(self.names)
        if self.rank < 2:
            return 1
        return self.node.shape[self.selecter.col]

    def GetValue(self, row, col):
        """ Callback which provides data to the Grid.

        row, col:   Integers giving row and column position (0-based).
        """
        try:
            # Scalar case
            if self.rank == 0:
                data = self.node[()]
                if self.names is None:
                    return self._format_value(data)
                return self._format_value(data[col])

            # 1D case
            if self.rank == 1:
                data = self.cache[row]
                if self.names is None:
                    return self._format_value(data)
                return self._format_value(data[self.names[col]])

            # ND case.  Watch out for compound mode!
            if self.names is not None:
                args = self.slicer.indices + (row,)
            else:
                l = []
                for x in range(self.rank):
                    if x == self.selecter.row:
                        l.append(row)
                    elif x == self.selecter.col:
                        l.append(col)
                    else:
                        idx = 0
                        for y in self.selecter.indices:
                            if y == x:
                                l.append(self.slicer.indices[idx])
                                break
                            idx = idx + 1
                args = tuple(l)

            data = self.cache[args]
            if self.names is None:
                return self._format_value(data)
            return self._format_value(data[self.names[col]])

        except Exception as e:
            return f"<Error: {e}>"

    def _format_value(self, value):
        """ Format a value for display in the grid """
        try:
            # Handle numpy scalar types
            # Only call .item() for 0-d arrays or arrays with exactly 1 element
            if hasattr(value, 'item') and hasattr(value, 'size'):
                if value.size == 1:
                    value = value.item()

            # Handle bytes
            if isinstance(value, bytes):
                return value.decode('utf-8', errors='replace')

            # Convert to string
            return str(value)
        except Exception as e:
            return f"<Format Error: {e}>"

    def GetRowLabelValue(self, row):
        """ Callback for row labels.

        Row number is used unless the data is scalar.
        """
        if self.rank == 0:
            return "Value"
        return str(row)

    def GetColLabelValue(self, col):
        """ Callback for column labels.

        Column number is used, except for scalar or 1D data, or if we're
        displaying field names in the columns.
        """
        if self.names is not None:
            return self.names[col]
        if self.rank == 0 or self.rank == 1:
            return "Value"
        return str(col)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Example data model which represents the file system.

Subclasses just two node types: Container and Array, representing
directories and files respectively.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import os.path as op

import numpy as np

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model


class Filesystem(compass_model.Store):
    """
        A "data store" represented by the file system.

        Keys are absolute paths on the local file system.
    """

    @staticmethod
    def plugin_name():
        return "Filesystem"

    @staticmethod
    def plugin_description():
        return "A plugin used to browse local files and folders."

    def __contains__(self, key):
        return op.exists(key)

    @property
    def url(self):
        return self._url

    @property
    def display_name(self):
        return "Local file system"

    @property
    def root(self):
        return self['/']

    @property
    def valid(self):
        return self._valid

    @staticmethod
    def can_handle(url):
        if url == "file://localhost":
            log.debug("able to handle %s? yes" % url)
            return True
        log.debug("able to handle %s? no" % url)
        return False

    def __init__(self, url):
        if not self.can_handle(url):
            raise ValueError(url)
        self._url = url
        self._valid = True

    def close(self):
        self._valid = False

    def get_parent(self, key):
        if key == "/":
            return None
        return self[op.dirname(key)]


class Directory(compass_model.Container):
    """
        Represents a directory in the filesystem.
    """

    class_kind = "Directory"

    @staticmethod
    def can_handle(store, key):
        return op.isdir(key)

    def __init__(self, store, key):
        self._store = store
        self._key = key
        try:
            self._names = os.listdir(key)
        except OSError:  # Permissions, etc.
            self._names = []

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        bn = op.basename(self.key)
        if len(bn) == 0:
            return "/"
        return bn

    @property
    def description(self):
        return 'Folder "%s" (%d members)' % (self.display_name, len(self))

    def __len__(self):
        return len(self._names)

    def __iter__(self):
        for name in self._names:
            key = op.join(self.key, name)
            yield self._store[key]

    def __getitem__(self, idx):
        key = op.join(self.key, self._names[idx])
        return self._store[key]


class File(compass_model.Array):
    """
        Represents a file (all loaded as an array of bytes)
    """

    class_kind = "File"

    @staticmethod
    def can_handle(store, key):
        return op.isfile(key)

    def __init__(self, store, key):
        self._store = store
        self._key = key

    @property
    def key(self):
        return self._key

    @property
    def store(self):
        return self._store

    @property
    def display_name(self):
        return op.basename(self.key)

    @property
    def description(self):
        return 'File "%s", size %d bytes' % (self.display_name, op.getsize(self.key))

    @property
    def shape(self):
        return (op.getsize(self.key),)

    @property
    def dtype(self):
        return np.dtype('u1')

    def __getitem__(self, args):
        try:
            with open(self.key, 'rb') as f:
                data = np.fromstring(f.read(), dtype='u1')
        except (OSError, IOError):
            data = np.zeros((len(self),), dtype='u1')

        return data[args]


Filesystem.push(File)
Filesystem.push(Directory)

compass_model.push(Filesystem)



"""
CLI wrapper for IOWarp Core binaries.

This module provides entry points that execute the bundled IOWarp binaries
with the correct library paths set.
"""

import os
import sys
import subprocess
from pathlib import Path


def get_package_paths():
    """Get the paths to the package lib and bin directories."""
    # Get the directory where this module is installed
    module_dir = Path(__file__).parent.resolve()

    lib_dir = module_dir / "lib"
    bin_dir = module_dir / "bin"

    return lib_dir, bin_dir


def run_binary(binary_name):
    """
    Run a bundled or system-installed binary with the correct library paths.

    Args:
        binary_name: Name of the binary to execute
    """
    lib_dir, bin_dir = get_package_paths()
    binary_path = bin_dir / binary_name

    # Check if binary is bundled in package directory
    if not binary_path.exists():
        # Fall back to system-installed binary (from source distribution)
        import shutil
        system_binary = shutil.which(binary_name)
        if system_binary:
            binary_path = Path(system_binary)
        else:
            print(f"Error: Binary '{binary_name}' not found", file=sys.stderr)
            print(f"  - Not bundled at: {bin_dir / binary_name}", file=sys.stderr)
            print(f"  - Not found in PATH", file=sys.stderr)
            print(f"Make sure iowarp-core is installed correctly.", file=sys.stderr)
            sys.exit(1)

    # Set up environment with library path
    env = os.environ.copy()

    # Build library path including conda environment if available
    lib_paths = []

    # Add bundled lib directory if it exists (for bundled binary distributions)
    if lib_dir.exists():
        lib_paths.append(str(lib_dir))

    # Add conda lib directory for dependencies (HDF5, MPI, etc.)
    conda_prefix = env.get("CONDA_PREFIX")
    if conda_prefix:
        conda_lib = Path(conda_prefix) / "lib"
        if conda_lib.exists():
            lib_paths.append(str(conda_lib))

    # Add lib directory to library path (only if we have paths to add)
    if lib_paths:
        if sys.platform.startswith("linux"):
            ld_library_path = env.get("LD_LIBRARY_PATH", "")
            new_paths = ":".join(lib_paths)
            if ld_library_path:
                env["LD_LIBRARY_PATH"] = f"{new_paths}:{ld_library_path}"
            else:
                env["LD_LIBRARY_PATH"] = new_paths
        elif sys.platform == "darwin":
            dyld_library_path = env.get("DYLD_LIBRARY_PATH", "")
            new_paths = ":".join(lib_paths)
            if dyld_library_path:
                env["DYLD_LIBRARY_PATH"] = f"{new_paths}:{dyld_library_path}"
            else:
                env["DYLD_LIBRARY_PATH"] = new_paths

    # Execute the binary with arguments passed from command line
    try:
        result = subprocess.run(
            [str(binary_path)] + sys.argv[1:],
            env=env,
            check=False
        )
        sys.exit(result.returncode)
    except Exception as e:
        print(f"Error executing {binary_name}: {e}", file=sys.stderr)
        sys.exit(1)


def wrp_cae_omni():
    """Entry point for wrp_cae_omni binary."""
    run_binary("wrp_cae_omni")


def chi_refresh_repo():
    """Entry point for chi_refresh_repo binary."""
    run_binary("chi_refresh_repo")


def chimaera_compose():
    """Entry point for chimaera_compose binary."""
    run_binary("chimaera_compose")


def chimaera_start_runtime():
    """Entry point for chimaera_start_runtime binary."""
    run_binary("chimaera_start_runtime")


def chimaera_stop_runtime():
    """Entry point for chimaera_stop_runtime binary."""
    run_binary("chimaera_stop_runtime")


def test_binary_assim():
    """Entry point for test_binary_assim binary."""
    run_binary("test_binary_assim")


def test_error_handling():
    """Entry point for test_error_handling binary."""
    run_binary("test_error_handling")


def test_hdf5_assim():
    """Entry point for test_hdf5_assim binary."""
    run_binary("test_hdf5_assim")


def test_range_assim():
    """Entry point for test_range_assim binary."""
    run_binary("test_range_assim")


# User-friendly aliases
def wrp_start():
    """Alias for chimaera_start_runtime (start IOWarp runtime)."""
    run_binary("chimaera_start_runtime")


def wrp_stop():
    """Alias for chimaera_stop_runtime (stop IOWarp runtime)."""
    run_binary("chimaera_stop_runtime")


def wrp_compose():
    """Alias for chimaera_compose (compose cluster configuration)."""
    run_binary("chimaera_compose")


def wrp_refresh():
    """Alias for chi_refresh_repo (refresh repository)."""
    run_binary("chi_refresh_repo")


def wrp_cae():
    """Alias for wrp_cae_omni (CAE OMNI processor)."""
    run_binary("wrp_cae_omni")



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Main module for HDFCompass.

Defines the App class, along with supporting infrastructure.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

# Must be at the top, to ensure we're the first to call matplotlib.use.
import matplotlib
matplotlib.use('WXAgg')

import sys
import wx
from wx.lib.pubsub import pub

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model
from hdf_compass import utils

from .events import ID_COMPASS_OPEN, EVT_COMPASS_OPEN
from . import container, array, geo_surface, geo_array, keyvalue, image, text
from .frame import BaseFrame, InitFrame, NodeFrame  # Import all needed frame classes

__version__ = utils.__version__


class CompassImageList(wx.ImageList):

    """
    A specialized type of image list, to support icons from Node subclasses.

    Instances of this class hold only square icons, of the size specified
    when created.  The appropriate icon index for a particular Node subclass is
    retrieved using get_index(nodeclass).

    Image addition and indexing is completely bootstrapped; there's no need
    to manually add or register Node classes with this class.  Just call
    get_index and the object will figure it out.
    """

    def __init__(self, size):
        """ Create a new list holding square icons of the given size. """
        wx.ImageList.__init__(self, size, size)
        self._indices = {}
        self._size = size

    def get_index(self, node_class):
        """ Retrieve an index appropriate for the given Node subclass. """

        if node_class not in self._indices:
            png = wx.Bitmap(node_class.icons[self._size], wx.BITMAP_TYPE_ANY)
            idx = self.Add(png)
            self._indices[node_class] = idx

        return self._indices[node_class]


class CompassApp(wx.App):

    """
    The main application object for HDFCompass.

    This mainly handles ID_COMPASS_OPEN events, which are requests to launch
    a new window viewing a particular node.  Also contains a dict of
    CompassImageLists, indexed by image width.
    """

    def __init__(self, redirect=False):
        """ Create a new compass application """
        super(CompassApp, self).__init__(redirect)
        self.imagelists = {}  # Initialize imagelists dictionary
        self.Bind(wx.EVT_MENU, self.on_about, id=wx.ID_ABOUT)
        self.Bind(wx.EVT_MENU, self.on_exit, id=wx.ID_EXIT)
        self.Bind(EVT_COMPASS_OPEN, self.on_compass_open)

    def OnInit(self):
        """ Initialize the application. """
        if not hasattr(self, 'imagelists'):
            self.imagelists = {}
        self.init_imagelists()
        self.frame = InitFrame()
        self.frame.Show()
        return True

    def init_imagelists(self):
        """ Initialize image lists for different icon sizes. """
        if not hasattr(self, 'imagelists'):
            self.imagelists = {}
        self.imagelists[16] = CompassImageList(16)
        self.imagelists[32] = CompassImageList(32)
        self.imagelists[64] = CompassImageList(64)

    def OnExit(self):
        """ Clean up when the application exits. """
        for il in self.imagelists.values():
            il.Destroy()
        return 0

    def on_compass_open(self, evt):
        """ A request has been made to open a node from somewhere in the GUI
        """
        log.debug(f"CompassApp.on_compass_open called with event: {evt}")
        # Get the node from the event
        node = evt.node
        log.debug(f"CompassApp.on_compass_open - node: {node.key}, type: {type(node)}")
        # Get the position from the event if available
        pos = getattr(evt, 'pos', None)
        log.debug(f"CompassApp.on_compass_open - calling open_node")
        # Open the node
        open_node(node, pos)

    def on_about(self, evt):
        """ Handle about menu item """
        info = wx.adv.AboutDialogInfo()
        info.SetName("HDF Compass")
        info.SetVersion(__version__)
        info.SetDescription("A viewer for HDF files")
        info.SetCopyright("Copyright (c) 2015-%d The HDF Group" % date.today().year)
        wx.adv.AboutBox(info)

    def on_exit(self, evt):
        """ Handle exit menu item """
        self.Exit()

    def MacOpenFile(self, filename):
        """ A file has been dropped onto the app icon """
        url = 'file://' + filename
        open_store(url)


def open_node(node, pos=None):
    """ Open a viewer frame appropriate for the given Node instance.

    node:   Node instance to open
    pos:    2-tuple with current window position (used to avoid overlap).
    """

    if pos is not None:
        # The thing we get from GetPosition isn't really a tuple, so
        # you have to manually cast entries to int or it silently fails.
        new_pos =(int(pos[0])+40, int(pos[1])+40)
    else:
        new_pos = None

    log.debug("Top-level open called for %s" % node)

    if isinstance(node, compass_model.Container):
        log.debug("Got Container")
        f = container.ContainerFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.GeoSurface):
        f = geo_surface.GeoSurfaceFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.GeoArray):
        f = geo_array.GeoArrayFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.Array):
        log.debug(f"Got Array - creating ArrayFrame for node: {node.key}")
        try:
            f = array.ArrayFrame(node, pos=new_pos)
            log.debug(f"ArrayFrame created successfully, showing window")
            f.Show()
        except Exception as e:
            log.exception(f"Error creating ArrayFrame: {e}")

    elif isinstance(node, compass_model.Xml):
        f = text.XmlFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.Text):
        f = text.TextFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.KeyValue):
        f = keyvalue.KeyValueFrame(node, pos=new_pos)
        f.Show()

    elif isinstance(node, compass_model.Image):
        f = image.ImageFrame(node, pos=pos)
        f.Show()
    else:
        pass


def open_store(url):
    """ Open the url using the first matching registered Store class.

    Returns True if the url was successfully opened, False otherwise.
    """
    stores = [x for x in compass_model.get_stores() if x.can_handle(url)]

    if len(stores) > 0:
        store_class = stores[0]
        log.debug("open_store: attempting to create store instance of type %s" % store_class.__name__)
        try:
            # Create an instance of the store
            store = store_class(url)
            log.debug("open_store: store instance created successfully")
            
            root = store.root
            log.debug("open_store: got root node: %s" % root)
            
            if root is not None:
                log.debug("open_store: calling open_node")
                open_node(root)
                log.debug("open_store: open_node completed successfully")
                return True
            else:
                log.error("open_store: root is None")
        except Exception as e:
            log.error("Error opening store: %s" % str(e))
            wx.MessageBox("Error opening file: %s" % str(e), "Error", wx.OK | wx.ICON_ERROR)
    else:
        wx.MessageBox("No store found that can handle this file type", "Error", wx.OK | wx.ICON_ERROR)
    return False


def can_open_store(url):
    """ checks url for first matching registered Store class.

    Returns True if the url can be successfully opened, False otherwise.
    """
    stores = [x for x in compass_model.get_stores() if x.can_handle(url)]

    if len(stores) > 0:
        instance = stores[0](url)
        return True

    return False


def load_plugins():
    """ Helper function that attempts to load all the plugins """

    # provide some info about the env in use
    import platform
    log.debug("Python %s %s on %s %s (%s)" % (platform.python_version(), platform.architecture()[0],
                                              platform.uname()[0], platform.uname()[2], platform.uname()[4]))
    import numpy
    log.debug("numpy %s" % numpy.__version__)
    log.debug("matplotlib %s" % matplotlib.__version__)
    log.debug("wxPython %s" % wx.__version__)

    from hdf_compass import compass_model

    try:
        from hdf_compass import filesystem_model
    except ImportError:
        log.warning("Filesystem plugin: NOT loaded")

    try:
        from hdf_compass import array_model
    except ImportError:
        log.warning("Array plugin: NOT loaded")

    try:
        from hdf_compass import hdf5_model
        import h5py
        log.debug("h5py %s" % h5py.__version__)
    except ImportError:
        log.warning("HDF5 plugin: NOT loaded")

    try:
        from hdf_compass import bag_model
        from lxml import etree
        # Try to import hydroffice.bag, but don't fail if it's not available
        try:
            from hydroffice import bag
            log.debug("hydroffice.bag %s" % bag.__version__)
        except ImportError:
            log.debug("Using h5py fallback for BAG files (hydroffice.bag not available)")
        log.debug("lxml %s (libxml %s, libxslt %s)"
                  % (etree.__version__, ".".join(str(i) for i in etree.LIBXML_VERSION),
                     ".".join(str(i) for i in etree.LIBXSLT_VERSION)))
    except (ImportError, OSError):
        log.warning("BAG plugin: NOT loaded")

    try:
        from hdf_compass import asc_model
    except ImportError:
        log.warning("Ascii grid plugin: NOT loaded")

    try:
        from hdf_compass import opendap_model
        from pydap import lib
        log.debug("pydap %s (protocol %s)"
                  % (".".join(str(i) for i in lib.__version__), ".".join(str(i) for i in lib.__dap__)))
    except ImportError:
        log.warning("Opendap plugin: NOT loaded")
    
    from hdf_compass import hdf5rest_model    
    try:
        from hdf_compass import hdf5rest_model
    except ImportError:
        log.warning("HDF5 REST plugin: NOT loaded")

    try:
        from hdf_compass import csv_model
        import pandas as pd
        log.debug("CSV plugin loaded (pandas %s)" % pd.__version__)
    except ImportError:
        log.warning("CSV plugin: NOT loaded (pandas not available)")

    try:
        from hdf_compass import parquet_model
        import pandas as pd
        import pyarrow
        log.debug("Parquet plugin loaded (pandas %s, pyarrow %s)" % (pd.__version__, pyarrow.__version__))
    except ImportError:
        log.warning("Parquet plugin: NOT loaded (pandas or pyarrow not available)")

    try:
        from hdf_compass import adios_model
        import adios
        log.debug("ADIOS %s" % adios.__version__)
    except ImportError:
        log.warning("ADIOS plugin: NOT loaded")


def run():
    """ Run HDFCompass.  Handles all command-line arguments, etc. """

    import os.path as op

    app = CompassApp(False)

    load_plugins()

    urls = sys.argv[1:]

    for url in urls:
        if "://" not in url:
            # Convert file path to URL format
            url = 'file://' + op.abspath(url)
        if not open_store(url):
            log.warning('Failed to open "%s"; no handlers' % url)

    # Remove duplicate frame creation since it's already created in OnInit
    if utils.is_darwin:
        wx.MenuBar.MacSetCommonMenuBar(app.frame.GetMenuBar())
    else:
        app.frame.Show()

    app.MainLoop()
